# coding=utf8
from typing import Union
from datetime import datetime


def upper_hex(byte: Union[str, bytes, int]) -> str:
    r"""
    >>> upper_hex('\x00')
    '00'
    >>> upper_hex(0x0)
    '00'
    >>> upper_hex(5)
    '05'
    >>> upper_hex(b'\x01')
    '01'
    >>> upper_hex('')
    Traceback (most recent call last):
    ...
    ValueError: expected single byte
    >>> upper_hex(b'')
    Traceback (most recent call last):
    ...
    ValueError: expected single byte
    >>> upper_hex('\x00\x01')
    Traceback (most recent call last):
    ...
    ValueError: expected single byte
    >>> upper_hex(b'\x00\x01')
    Traceback (most recent call last):
    ...
    ValueError: expected single byte
    """
    if isinstance(byte, (str, bytes)):
        if len(byte) != 1:
            raise ValueError('expected single byte')
        if isinstance(byte, str):
            byte = ord(byte)
        elif isinstance(byte, bytes):
            byte = byte[0]
    return '%02X' % byte


def hex_str(byte_string, splitter='') -> str:
    r"""
    >>> hex_str('Python', ' ')
    '50 79 74 68 6F 6E'
    >>> hex_str('\x00\xa1\xb2', ' ')
    '00 A1 B2'
    >>> hex_str([1, 2, 3, 5, 8, 13], ' ')
    '01 02 03 05 08 0D'
    """
    return splitter.join(upper_hex(c) for c in byte_string)


def to_datetime(byte_string, from_format, to_format) -> str:
    r"""
    >>> to_datetime(b'\x16\x01\x23', '%d%m%y', '%Y-%m-%d')
    '2023-01-16'
    >>> to_datetime(b'\x01\x13\x57\x43\x16\x01\x23', '01%H%M%S%d%m%y', '%Y-%m-%d %H:%M:%S')
    '2023-01-16 13:57:43'
    """
    from_string = hex_str(byte_string)
    parsed_datetime = datetime.strptime(from_string, from_format).date()
    return parsed_datetime.strftime(to_format)


def chunk_string(string, length) -> iter:
    return (string[0 + i:length + i] for i in range(0, len(string), length))



# coding=utf8
from mercury_base import Meter
from mercury_base.utils import chunk_string, hex_str, to_datetime


def get_group_address(meter: Meter):
    data = meter.send_command(0x20)
    return hex_str(data)


def get_datetime(meter: Meter):
    data = meter.send_command(0x21)
    return to_datetime(data, '01%H%M%S%d%m%y', '%Y-%m-%d %H:%M:%S')


def get_power_limit(meter: Meter):
    data = meter.send_command(0x22)
    return int(hex_str(data)) / 100.


def get_month_energy_limit(meter: Meter):
    data = meter.send_command(0x23)
    return int(hex_str(data)) / 100.


def get_is_seasonal_time(meter: Meter):
    data = meter.send_command(0x24)
    return data != 0x0


def get_time_correction(meter: Meter):
    data = meter.send_command(0x25)
    return data[0]


def get_power(meter: Meter):
    data = meter.send_command(0x26)
    return int(hex_str(data)) / 100.


def get_energy_accumulators(meter: Meter):
    data = meter.send_command(0x27)
    return [
        int(hex_str(data[0:4])) / 100.,
        int(hex_str(data[4:8])) / 100.,
        int(hex_str(data[8:12])) / 100.,
        int(hex_str(data[12:16])) / 100.,
    ]


def get_firmware_info(meter: Meter):
    data = meter.send_command(0x28)
    return {
        'version': str(int(hex_str(data[0:2]))) + '.' + str(int(hex_str(data[0:2]))),
        'date': to_datetime(data[2:6], '%d%m%y', '%Y-%m-%d'),
    }


def get_battery_voltage(meter: Meter):
    data = meter.send_command(0x29)
    return int(hex_str(data)) / 100.


def get_display_filters(meter: Meter):
    data = meter.send_command(0x2A)
    return format(list(data)[0], '0>8b')


def get_last_stop_datetime(meter: Meter):
    data = meter.send_command(0x2B)
    return to_datetime(data, '01%H%M%S%d%m%y', '%Y-%m-%d %H:%M:%S')


def get_last_start_datetime(meter: Meter) -> str:
    data = meter.send_command(0x2C)
    return to_datetime(data, '01%H%M%S%d%m%y', '%Y-%m-%d %H:%M:%S')


def get_output_optocoupler_function(meter: Meter) -> int:
    data = meter.send_command(0x2D)
    return int(hex_str(data))


def get_tariffs_count(meter: Meter) -> int:
    data = meter.send_command(0x2E)
    return int(hex_str(data))


def get_serial_number(meter: Meter) -> int:
    data = meter.send_command(0x2F)
    serial_number = hex_str(data)
    return int(serial_number, 16)


def get_holidays(meter: Meter) -> list:
    holidays = []
    for part in [0, 1]:
        data = meter.send_command(0x30, part)
        for chunk in chunk_string(data, 2):
            chunk_data = hex_str(chunk)
            if chunk_data != 'FFFF':
                holidays.append(to_datetime(chunk, '%d%m', '%m-%d'))
    return holidays


def get_vcp(meter: Meter) -> dict:
    data = meter.send_command(0x63)
    return {
        'voltage': int(hex_str(data[0:2])) / 10.,
        'current': int(hex_str(data[2:4])) / 100.,
        'power': int(hex_str(data[4:7])) / 1000.,
    }


def get_tariff(meter: Meter) -> int:
    data = meter.send_command(0x60)
    return int(hex_str(data))


def set_display_filters(meter: Meter, params) -> bool:
    """
    TODO
    """
    return True



# coding=utf8
from mercury_base import ADDRESS_FMT
from mercury_base.mercury_v1 import commands
from struct import unpack


def extract_address(message: bytes) -> int:
    address = unpack(ADDRESS_FMT, message[:4])[0]
    return address


def extract_command(message: bytes) -> bytes:
    command = list(message[4:5])[0]
    return command


def extract_data(message: bytes) -> list[bytes]:
    data = list(message[5:-2])
    return data



# coding=utf8
"""
Toolkit for communicating with Incotex Mercury meters via RS485/CAN bus
Copyright (c) 2023 webtoucher
Distributed under the BSD 3-Clause license. See LICENSE for more info.
"""
import importlib
import serial
import time

from mercury_base.utils import hex_str
from modbus_crc import add_crc, check_crc
from struct import pack
from typing import Final


class ConnectError(Exception):
    pass


class CommunicationError(Exception):
    pass


class UnexpectedAddress(CommunicationError):
    pass


class UnexpectedCommand(CommunicationError):
    pass


class CheckSumError(CommunicationError):
    pass


ADDRESS_FMT: Final[str] = '!I'  # unsigned integer in network order

TYPE_200: Final[str] = '200'
TYPE_201: Final[str] = '201'
TYPE_203: Final[str] = '203'
TYPE_206: Final[str] = '206'
TYPE_203_2TD: Final[str] = '203.2TD'
TYPE_204: Final[str] = '204'
TYPE_208: Final[str] = '208'
TYPE_230: Final[str] = '230'
TYPE_231: Final[str] = '231'
TYPE_234: Final[str] = '234'
TYPE_236: Final[str] = '236'
TYPE_238: Final[str] = '238'

TYPES_V1: Final[list] = [
    TYPE_200,
    TYPE_201,
    TYPE_203,
    TYPE_206,
]

TYPES_V2: Final[list] = [
    TYPE_203_2TD,
    TYPE_204,
    TYPE_208,
    TYPE_230,
    TYPE_231,
    TYPE_234,
    TYPE_236,
    TYPE_238,
]


class Meter(object):
    def __init__(self, meter_type: str, address: int, port: str, logger=None,
                 baudrate=9600, parity=serial.PARITY_NONE, bytesize=8, stopbits=1, timeout=0.05):
        self.__type = meter_type
        self.__address = address
        self.__port = port
        self.__logger = logger
        self.__serial_number = None

        if meter_type in TYPES_V1:
            module = 'mercury_v1'
        elif meter_type in TYPES_V2:
            raise ConnectError('Meter Mercury %s is not supported yet' % meter_type)
        else:
            raise ConnectError('Meter Mercury %s is not supported' % meter_type)

        self.__driver = importlib.import_module('.%s' % module, __package__)
        self.__connection = serial.Serial(port=port, baudrate=baudrate, parity=parity,
                                          bytesize=bytesize, stopbits=stopbits, timeout=timeout)
        self.__serial_number = self.command('get_serial_number')
        if not self.__serial_number:
            raise ConnectError('Meter at address %s did not respond' % self.__address)
        if self.__logger:
            self.__logger.info('Meter with serial number %s is connected', self.__serial_number)

    @property
    def type(self) -> str:
        return self.__type

    @property
    def serial_number(self) -> int:
        return self.__serial_number

    def has_command(self, command: str) -> bool:
        """ Check if command exists """
        return hasattr(self.__driver.commands, command)

    def command(self, command: str, *params) -> any:
        """ Send command to the meter """
        command_method = getattr(self.__driver.commands, command)
        return command_method(self, *params)

    def test_package(self, package: bytes) -> bool:
        """ Check is package correct for this meter """
        try:
            return self.__address == self.__driver.extract_address(package)
        except Exception:
            return False

    def send_package(self, package: bytes, attempts=5) -> bytes:
        """ Send raw data to the meter """
        buffer_size = 1024
        answer = None
        if not check_crc(package):
            raise CheckSumError('Outgoing package is incorrect')
        self.__connection.write(package)
        if self.__logger:
            self.__logger.debug('--> [%s]\t%s', self.__serial_number or 'new meter', hex_str(package, ' '))
        while not answer and attempts:
            attempts -= 1
            time.sleep(0.1)
            answer = self.__connection.read(buffer_size)
            if answer:
                if not check_crc(answer):
                    raise CheckSumError('Incoming package is incorrect')
                address = self.__driver.extract_address(answer)
                if address != self.__driver.extract_address(package):
                    raise UnexpectedAddress(address)
                command = self.__driver.extract_command(answer)
                if command != self.__driver.extract_command(package):
                    raise UnexpectedCommand(command)
                if self.__logger:
                    self.__logger.debug('<-- [%s]\t%s', self.__serial_number or 'new meter', hex_str(answer, ' '))
        return answer

    def send_command(self, *params, attempts=5) -> bytes:
        address = pack(ADDRESS_FMT, self.__address)
        package = add_crc(address + bytes(params))
        received_package = self.send_package(package, attempts=attempts)
        return self.__driver.extract_data(received_package)


if __name__ == '__main__':
    pass



from setuptools import setup

setup(name='mercury-base',
      version='1.0a3',
      url='https://github.com/webtoucher/mercury-base',
      license='BSD-3-Clause',
      author='Alexey Kuznetsov',
      author_email='mirakuru@webtoucher.ru',
      description='Toolkit for communicating with Incotex Mercury meters via RS485/CAN bus',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Topic :: Terminals :: Serial',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 3',
          'Operating System :: Unix',
      ],
      packages=['mercury_base', 'mercury_base.mercury_v1'],
      long_description=open('README.md', encoding='utf-8').read(),
      long_description_content_type='text/markdown',
      zip_safe=False)
